import math
import time
from datetime import datetime, timezone

from autotraders.map.waypoint import Waypoint

class WaypointQuery:
    def __init__(self, args):
        self.data = args

    def query(self, ship, session):
        if type(self.data["destination"]) is str:
            return self.data["destination"]
        else:
            choices, _ = Waypoint.all(ship.nav.location.system, session)
            ship_current_waypoint = [choice for choice in choices if choice.symbol == ship.nav.location][0]
            tmp = []
            for choice in choices:
                if "trait" in self.data["destination"]:
                    if len([trait for trait in choice.traits if
                            trait.symbol == self.data["destination"]["trait"]]) != 0:
                        tmp.append(choice)
                if "type" in self.data["destination"]:
                    if choice.waypoint_type == self.data["destination"]["type"]:
                        tmp.append(choice)
            choices = tmp
            if len(choices) == 0:
                raise Exception("0 possible waypoints")
            if "selection" not in self.data["destination"] or self.data["destination"][
                "selection"] == "NEAREST":
                accepted = choices[0]
                best = 1000000
                for choice in choices:
                    if choice.symbol == ship.nav.location:
                        accepted = choice
                        best = -1
                    distance = math.sqrt(
                        ((choice.x - ship_current_waypoint.x) ** 2) + ((choice.y - ship_current_waypoint.y) ** 2))
                    if distance < best:
                        best = distance
                        accepted = choice
            else:
                accepted = choices[0]
            return accepted


class Navigate:
    def __init__(self, args):
        self.data = args
        self.query = WaypointQuery(args)

    def execute(self, ship, session):
        accepted = self.query.query(ship, session)
        if str(accepted.symbol) != str(ship.nav.location):
            ship.navigate(str(accepted.symbol))
        time.sleep(1)
        ship.update()
        if ship.nav.status == "IN_TRANSIT":
            time.sleep((ship.nav.route.arrival - datetime.now(timezone.utc)).seconds)



import logging

from autotraders.ship import Ship


class ShipCommand:
    def __init__(self, command, args):
        if type(args) is not dict:
            args = {}
        self.command = command
        self.fail_fast = args.get("fail-fast", True)

    def execute(self, ship: Ship):
        if self.command == "dock":
            ship.dock()
        elif self.command == "orbit":
            ship.orbit()
        elif self.command == "refuel":
            try:
                ship.refuel()
            except IOError as e:
                if self.fail_fast:
                    raise e
                else:
                    logging.warning(self.command + " failed to execute")
        elif self.command == "extract":
            try:
                ship.refuel()
            except IOError as e:
                if self.fail_fast:
                    raise e
                else:
                    logging.warning(self.command + " failed to execute")



import logging
import time
from threading import Thread

import yaml
from autotraders.agent import Agent
from autotraders.ship import Ship

import secret
from spacelang.step import Step


class Action:
    def __init__(self, name, data):
        self.name = name
        self.actions = []
        for d in data["steps"]:
            if type(d) == str:
                self.actions.append(Step(d, None))
            else:
                self.actions.append(Step(list(d.keys())[0], d[list(d.keys())[0]]))

    def execute(self, ship, events, session):
        for action in self.actions:
            action.execute(ship, events, session)

    def run(self, ships, events, session):
        for ship in ships:
            self.execute(ship, events, session)
            time.sleep(1)


class Trigger:
    def __init__(self, name, data):
        self.name = name  # TODO: Concurrency support
        self.steps = {}
        for group in data:
            for d in data[group]["steps"]:
                self.steps[group] = []
                if type(d) == str:
                    self.steps[group].append(Step(d, None))  # TODO: Move to step class
                else:
                    self.steps[group].append(Step(list(d.keys())[0], d[list(d.keys())[0]]))

    def run(self, ships, events, session):
        for group in self.steps:
            ship_group = ships[group]
            for ship in ship_group:
                for step in self.steps[group]:
                    step.execute(ship, events, session)


class File:
    def __init__(self, data):
        self.ship_groups = data["ships"]
        self.events = [Action(d, data["actions"][d]) for d in data["actions"]]
        self.triggers = [Trigger(d, data["triggers"][d]) for d in data["triggers"]]

    def run(self, session):
        logging.info("Initializing state ...")
        agent = Agent(session)
        logging.info("Agent: " + agent.symbol)
        ships = {}
        for group in self.ship_groups:
            ships[group] = [Ship(ship, session) for ship in self.ship_groups[group]]
            time.sleep(0.5)
        contains_onstart = len([trigger for trigger in self.triggers if trigger.name == "on_start"]) == 1
        if contains_onstart:
            on_start = [trigger for trigger in self.triggers if trigger.name == "on_start"][0]
            logging.info("Processing trigger on_start")
            thread = Thread(target=on_start.run, name="OnStart", args=(ships, self.events, session))
            thread.start()
        while True:
            logging.debug("Checking triggers ...")
            # agent.update()
            time.sleep(10)  # TODO: Actually check triggers


def load_text(stream):
    data = yaml.load(stream, Loader=yaml.Loader)
    return data


logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.basicConfig(format='[%(threadName)s] %(levelname)s: %(message)s', level=logging.DEBUG)



import logging
import time
from autotraders.ship import Ship

from spacelang.step.navigate import Navigate
from spacelang.step.ship_command import ShipCommand


class Step:
    def __init__(self, command, data):
        self.command = command
        self.args = data
        if self.command == "navigate":
            self.inner = Navigate(self.args)
        elif self.command in ["dock", "orbit", "refuel", "extract"]:
            self.inner = ShipCommand(self.command, self.args)

    def action(self, ship, events, session):
        repeat_interval = -1
        if type(self.args) is str:
            name = self.args
            repeat = False
        else:
            name = self.args["name"]
            repeat = "repeat" in self.args
            if repeat and "interval" in self.args["repeat"]:
                repeat_interval = self.args["repeat"]["interval"]
        for event in events:
            if event.name == name:
                if repeat:
                    if repeat_interval != -1:
                        while True:
                            event.execute(ship, events, session)
                            time.sleep(repeat_interval)
                else:
                    event.execute(ship, events, session)

    def execute(self, ship: Ship, events, session):
        logging.debug(self.command)
        if self.command == "navigate":
            self.inner.execute(ship, session)
        elif self.command in ["dock", "orbit", "refuel", "extract"]:
            self.inner.execute(ship)
        elif self.command == "action":
            self.action(ship, events, session)
        elif self.command == "refine":
            ship.refine(self.args)
        elif self.command == "sell":
            assert type(self.args) is dict
            ship.sell(self.args["symbol"], self.args.get("quantity", 1))
        elif self.command == "buy":
            assert type(self.args) is dict
            ship.buy(self.args["symbol"], self.args.get("quantity", 1))
        elif self.command == "sellall":
            exclude = []
            if type(self.args) is dict and "exclude" in self.args:
                exclude = self.args["exclude"]
            for i in ship.cargo.inventory:
                if i not in exclude:
                    ship.sell(i, ship.cargo.inventory[i])
                    time.sleep(0.1)
        elif self.command == "sleep":
            if type(self.args) is int:
                time.sleep(self.args)
            else:
                raise NotImplementedError("TODO")
        else:
            raise NotImplementedError("TODO")
        time.sleep(1)
